
program adaline(input,output);
{Simulation of Bernard Widrow's Adaptive Linear Element.
Program Copyright 1987 by Wesley R. Elsberry & Diane J. Blackwood.
Released for non-commercial use and distribution.  Program procedures
may be modified for use, as long as authors are properly credited.
Commercial distribution is prohibited.  See the ADAL.HLP file for
valid commands.}

const
 ESC = ^[;
 max_size = 16;

type
  active_memory_ = 1..max_size;
  full_line_ = string[80];
  memister_ = real;
  memaray_ = array[1..max_size,1..max_size] of
               record
                 mem : memister_;
                 w : integer;
               end;
  sa_ptr_ = ^sense_array_;
  sense_array_ = record
                   m : memaray_;
                   bu : memister_;
                   actmem : active_memory_;
                 end;
  page_ = array[1..24] of full_line_;
  row_ =  1..24;
  col_ =  1..80;

var
  sa1     : sense_array_; { sense array }
  actmem  : byte;         { square root of size of active memory }
  ii      : integer;      { index }
  jj      : integer;      { index }
  done    : boolean;      { boolean for main WHILE loop        }
  use_f   : boolean;      { boolean for selecting input source }
  log_it  : boolean;      { boolean for logging transactions   }
  ioerr   : boolean;      { flags I/O errors                   }
  do_up   : boolean;      { boolean for update at end of case  }
  inchar  : char;         { keyboard input variable            }
  infile  : text;         { input file handle                  }
  inf     : text;         { another input file handle          }
  outfile : text;         { file handle for logfile            }
  help    : page_;        { a text screen for help messages    }
  fname   : string[50];   { filename variable                  }
  ofname  : string[50];   { another filename variable          }
  row     : row_;         { subrange variable for cursor pos.  }
  col     : col_;         { subrange variable for cursor pos.  }
  mval    : real;         { the current meter value            }
  noisy   : boolean;      { add noise to system?  T/F          }
  noisev  : real;         { variance of noise                  }

{$i iocheck.pas }

function tirandom(mean,variance:real):real;
var u1,u2,x : real;
begin
 u1 := random;
 u2 := random;
 x := sqrt(-2*ln(u1))*cos(2*pi*u2);
 x := variance*x + mean;
 tirandom := x;
end;

function increment_sa
  (c_val,g_val : real;  { current value and goal value }
   var sa : sense_array_):boolean;
var
  delta_val,i_val : real;
  ii, jj : integer;
  it : boolean;
begin
  delta_val := g_val - c_val;
  i_val := delta_val/(sa.actmem*sa.actmem+1);
  for ii := 1 to sa.actmem do
    for jj := 1 to sa.actmem do
      begin
        sa.m[ii,jj].mem := sa.m[ii,jj].mem + (i_val * sa.m[ii,jj].w);
        if noisy then sa.m[ii,jj].mem := sa.m[ii,jj].mem + tirandom(0.0,noisev);
      end;
  sa.bu := sa.bu + i_val;
  if (delta_val > 0.01) or (delta_val < -0.01) then
    it := false
  else
    it := true;
  increment_sa := it;
end;

procedure analog_meter(rr : real);
var ii, jj, kk : integer;
begin
  if abs(rr) > 10 then halt;
  ii := round(5*rr);
  if ii < -5 then ii := -5;
  if ii > 5 then ii := 5;
  {position cursor at 0 spot, positive side}
  gotoxy(1,10); write('     |     ');
  if ii < 0 then
    begin
      kk := abs(ii);
      for jj := 1 to kk do
        begin
          gotoxy((6-jj),10);
          write('-');
        end;
    end
  else
    begin
      gotoxy(7,10);
      for jj := 1 to ii do
        write('+');
    end;
end;

function meter(sa : sense_array_) : real;
var
  isum : real;
  ii, jj :integer;
begin
  mval := 0.0;
  for ii := 1 to sa.actmem do  {Sum inputs times weights}
    for jj := 1 to sa.actmem do
      begin
        mval := mval + (sa.m[ii,jj].w * sa.m[ii,jj].mem);
      end;
  mval := mval + sa.bu;
  meter := mval;
end;

procedure set_meter(gval : real; var sa : sense_array_);
{Change weights until meter gives desired value}
begin
  while not increment_sa(meter(sa),gval,sa) do ;
end;

procedure reset_sa(var sa : sense_array_); {Restart sense array}
var ii , jj : integer; xx : real;
begin
  xx := 0.0;
  for ii := 1 to sa.actmem do
    for jj := 1 to sa.actmem do
      begin
        sa.m[ii,jj].mem := xx;
        sa.m[ii,jj].w := -1;
      end;
  sa.bu := xx;
  set_meter(0.0,sa);
end;

procedure plot_pat(x,y :integer; var sa :sense_array_);
begin
        row := 3 + (y-1)*2; col := 26 + (x-1)*4;
        gotoxy(col,row);
        if sa.m[x,y].w = 1 then
          write('XX')
        else if sa.m[x,y].w = -1 then
          write('--')
        else
          begin
            sa.m[x,y].w := -1;
            write('--');
          end;
end;

procedure update(sa : sense_array_); {Refresh screen display}
var ii, jj : integer;
begin
  gotoxy(1,23);  write('<?> for help':15,' ':65);
  gotoxy(8,2);   write('METER');
  gotoxy(6,3);   write('---------');
  gotoxy(6,5);   write(meter(sa):7:3);
  gotoxy(6,7);   write('---------');
  analog_meter(mval);
  gotoxy(1,11);  write('^    ^    ^');
  gotoxy(1,12);  write('-    0    +');
  gotoxy(30,1);  write('PATTERN');
  gotoxy(30,12); write('WEIGHTS');
  for ii := 1 to sa.actmem do
    for jj := 1 to sa.actmem do
      begin
        plot_pat(ii,jj,sa);
        row := 12 + 2*jj; col := 24 + (ii-1)*7;
        gotoxy(col,row); write(sa.m[ii,jj].mem:5:2);
      end;
  gotoxy(8,14); write('BIAS UNIT');
  gotoxy(8,16); write(sa.bu:6:3);
  gotoxy(1,20);
end;

procedure help_user; {List a brief help page}
var ii : integer;
begin
  clrscr;
  for ii := 1 to 23 do
    writeln(help[ii]);
  write(help[24]);
  read(kbd,inchar);
  clrscr;
end;

procedure save_weights(sa : sense_array_);
var ii, jj : integer; wt_file : file of sense_array_;
begin
  repeat
    gotoxy(1,23); write(' ':80);
    gotoxy(1,23); write('File to save weights to: ');
    readln(fname);
    assign(wt_file,fname); {$I-}
    rewrite(wt_file); iocheck;    {$I+}
  until not ioerr;
  write(wt_file,sa);
  close(wt_file);
end;

procedure load_weights(var  sa : sense_array_);
var ii, jj : integer; wt_file : file of sense_array_;
sa_new : sense_array_;
begin
  repeat
    gotoxy(1,23); write(' ':80);
    gotoxy(1,23); write('File to get replacement weights from: ');
    readln(fname);
    assign(wt_file,fname); {$I-}
    reset(wt_file); iocheck;    {$I+}
  until not ioerr;
  read(wt_file,sa);
  close(wt_file);
end;

procedure add_weights(var sa : sense_array_);
var ii, jj : integer; wt_file : file of sense_array_; sa_add : sense_array_;
begin
  repeat
    gotoxy(1,23); write(' ':80);
    gotoxy(1,23); write('File to get weights from for adding: ');
    readln(fname);
    assign(wt_file,fname); {$I-}
    reset(wt_file); iocheck;    {$I+}
  until not ioerr;
  read(wt_file,sa_add);
  close(wt_file);
  for ii := 1 to sa.actmem do
    for jj := 1 to sa.actmem do
      sa.m[ii,jj].mem := sa.m[ii,jj].mem + sa_add.m[ii,jj].mem;
  sa.bu := sa.bu + sa_add.bu;
end;

procedure change_pat(x,y : integer; var sa : sense_array_);
begin
  sa.m[x,y].w := sa.m[x,y].w * -1;
  plot_pat(x,y,sa);
  gotoxy(6,5);   write(meter(sa):8:4);
  analog_meter(mval);
end;

begin
  assign(infile,'adal.hlp'); reset(infile);
  for ii := 1 to 24 do
    readln(infile,help[ii]);
  close(infile);
  help_user;
  {Activate ADALINE units, set active memory}

  sa1.actmem := 4;

  {initialize ADALINES}
  noisy := false;
  noisev := 0.005;
  reset_sa(sa1);
  update(sa1);
  done := false;
  log_it := false;
  use_f := false;
  while not done do
    begin
      {poll}
      do_up := false;
      gotoxy(1,24); write('READY: ');
      if eof(infile) then use_f := false;
      if use_f then
        read(infile,inchar)
      else
        read(kbd,inchar);
      if log_it then
        write(outfile,inchar);
      gotoxy(1,24);
      if inchar >= ' ' then
        write(inchar:7,' ':33);

      {act}
      case inchar of
        '1'     : change_pat(1,1,sa1);
        '2'     : change_pat(2,1,sa1);
        '3'     : change_pat(3,1,sa1);
        '4'     : change_pat(4,1,sa1);
        'q','Q' : change_pat(1,2,sa1);
        'w','W' : change_pat(2,2,sa1);
        'e','E' : change_pat(3,2,sa1);
        'r','R' : change_pat(4,2,sa1);
        'a','A' : change_pat(1,3,sa1);
        's','S' : change_pat(2,3,sa1);
        'd','D' : change_pat(3,3,sa1);
        'f','F' : change_pat(4,3,sa1);
        'z','Z' : change_pat(1,4,sa1);
        'x','X' : change_pat(2,4,sa1);
        'c','C' : change_pat(3,4,sa1);
        'v','V' : change_pat(4,4,sa1);
        '+' : done := true;
        '0' : begin
                gotoxy(1,24); write('Reset destroys weights. Are you sure?');
                if use_f then
                  read(infile,inchar)
                else
                  read(kbd,inchar);
                if inchar in ['Y','y'] then
                  reset_sa(sa1);
                clrscr;
                do_up := true;
              end;
        'y','Y' : begin
                    set_meter(1.0,sa1);
                    do_up := true;
                  end;
        'n','N' : begin
                    set_meter(-1.0,sa1);
                    do_up := true;
                  end;
        '?' : begin help_user; do_up := true; end;
        '9' : begin        {will need to subroutine and parameterize this code}
                for ii := 1 to sa1.actmem do
                  begin
                    for jj := 1 to sa1.actmem do
                      begin
                        sa1.m[ii,jj].w := -1;
                        plot_pat(ii,jj,sa1);
                      end;
                  end;
                gotoxy(6,5);   write(meter(sa1):8:4); analog_meter(mval);
              end;
        'l','L' : begin {learn from file}
                   repeat
                    gotoxy(1,23); write(' ':80);
                    gotoxy(1,23); write('Filename to learn from: ');
                    readln(fname);
                    close(infile);
                    assign(infile,fname); {$I-}
                    reset(infile); iocheck;    {$I+}
                    until not ioerr;
                    use_f := true;
                  end;
         'o','O' : begin
                     if log_it then
                       begin
                         gotoxy(1,23); write('Must close current logfile first');
                       end
                     else
                       begin
                        repeat
                         gotoxy(1,23); clreol;
                         gotoxy(1,23); write('Log Filename:');
                         readln(ofname);
                         gotoxy(1,24);clreol;
                         assign(outfile,ofname);
{$I-}                         rewrite(outfile); iocheck; {$I+}
                        until not ioerr;
                         log_it := true;
                       end;
                   end;
          'p','P' : begin
                      flush(outfile);
                      close(outfile);
                      log_it := false;
                    end;
          'U','u' : save_weights(sa1);
          'J','j' : begin
                      load_weights(sa1);
                      do_up := true;
                    end;
          'M','m' : begin
                      add_weights(sa1);
                      do_up := true;
                    end;
          '.'     : noisy := not noisy;
          '<'     : noisev := noisev/1.4;
          '>'     : noisev := noisev * 1.4;
          '=' : begin
                 gotoxy(20,23);
                 write('Version 0.51, Copyright 1987 W.R.Elsberry & D.J.Blackwood':60);
                end;
        else ;
      end;
      if do_up then update(sa1);
    end;
  clrscr;
  close(infile);
  flush(outfile); close(outfile);
end.


version 0.30  Added file logging and learning to Widrow ADALINE routines.
version 0.35  Added weight saving and restoring routines.  Deleted
              boolean parameters from METER, RESET, and SETMETER.
version 0.40  Removed fidget value and increment.  Added ADD_WEIGHTS procedure.
              Removed Zenith specific terminal sequences.
version 0.45  Revised case statement to speed response.  Changed prompts to
              be more informative.  Removed zero meter command (identical effect
              as reset).
version 0.50  Parameterized sense array (clearing way for MADALINE).  Corrected
              clutter of old prompts.  Added analog meter function.
version 0.51  Added noise to system.
